#!/usr/bin/env python3
#
# Copyright (c) 2015-2019 RunSafe Security Inc.
# All rights reserved.

from bisect import bisect_left
from copy import copy
import os
import struct
import sys

if 'LFR_PYTHONPATH' in os.environ:
    sys.path.append(os.environ['LFR_PYTHONPATH'])

from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_RELOC_TYPE_i386, ENUM_RELOC_TYPE_x64, ENUM_RELOC_TYPE_ARM
from parse_textrap import TargetArch, TrapWriter, ElfTrapParser, TrapHeader, TrapRecordV2, TrapRelocV2


# Adapted from http://graphics.stanford.edu/~seander/bithacks.html#VariableSignExtend
# In the public domain
def signextend(x, bits):
    m = 1 << (bits - 1)
    return (x ^ m) - m


def bit(x, b):
    return (x >> b) & 1


def bit_select(mask, x, y):
    return (x & ~mask) | (y & mask)


def align(x, y):
    return x & ~(y - 1)


# Relocation implementation based on RandoLib/posix/<arch>/Relocation.cpp
class Relocation:
    def __init__(self, reloc, builder):
        self.reloc = reloc
        self.builder = builder


class RelocationArm(Relocation):
    def __init__(self, reloc, builder):
        super().__init__(reloc, builder)

    def get_target_ptr(self):
        ty = self.reloc.type

        address = self.reloc.addr
        reloc_contents = self.builder.read_vaddr_typed(self.reloc.addr, '<I')
        signed_contents = self.builder.read_vaddr_typed(self.reloc.addr, '<i')
        if reloc_contents is None or signed_contents is None:
            return None

        def arm_bl_isX(ins):
            return (ins & 0xfe000000) == 0xfa000000

        def thm_bl_isX(ins):
            return bit(ins, 28) ^ 1

        def imm24_get(ins):
            return signextend((ins & 0xffffff) << 2, 26)

        def thm_imm24_get(ins):
            S = bit(ins, 10)
            I1 = bit(ins, 29) ^ S ^ 1
            I2 = bit(ins, 27) ^ S ^ 1
            value = ((S << 24) | (I1 << 23) | (I2 << 22) |
                     ((ins & 0x3ff) << 12) | ((ins & 0x7ff0000) >> 15))
            return signextend(value, 25)

        def thm_imm19_get(ins):
            S = bit(ins, 10)
            J1 = bit(ins, 29)
            J2 = bit(ins, 27)
            value = ((S << 20) | (J2 << 19) | (J1 << 18) |
                     ((ins & 0x3f) << 12) | ((ins & 0x7ff0000) >> 15))
            return signextend(value, 21)

        def B_T1_get(ins):
            return signextend((ins & 0x7ff) << 1, 12)

        def B_T2_get(ins, dest):
            return bit_select(0xff, ins, (dest & 0x1fe) >> 1)

        if ty in [ENUM_RELOC_TYPE_ARM['R_ARM_ABS32'], ENUM_RELOC_TYPE_ARM['R_ARM_TARGET1']]:
            return reloc_contents & 0xfffffffe
        if ty in [ENUM_RELOC_TYPE_ARM['R_ARM_REL32'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_BASE_PREL'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_GOT_PREL'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_TARGET2'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_TLS_GD32'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_TLS_LDM32'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_TLS_IE32']]:
            return address + signed_contents - self.reloc.addend
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_PREL31']:
            return signextend(reloc_contents & 0x7fffffff, 31) + address - self.reloc.addend
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_GOTOFF32']:
            if self.builder.got_address is None:
                return None
            else:
                return self.builder.got_address + signed_contents - self.reloc.addend
        # Instruction relocs
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_CALL']:
            if arm_bl_isX(reloc_contents):
                return imm24_get(reloc_contents) + address + 8 + ((reloc_contents >> 23) & 0x2)
        if ty in [ENUM_RELOC_TYPE_ARM['R_ARM_CALL'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_JUMP24'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_PLT32']]:
            return imm24_get(reloc_contents) + address + 8
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_V4BX']:
            # The linker will translate the instruction at this relocation
            # to either B <veneer> or MOV PC, <reg>
            if (reloc_contents & 0x0f000000) == 0x0a000000:
                return imm24_get(reloc_contents) + address + 8;
            return None;
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_THM_CALL'] and thm_bl_isX(reloc_contents):
            return thm_imm24_get(reloc_contents) + align(address + 4, 4)
        if ty in [ENUM_RELOC_TYPE_ARM['R_ARM_THM_CALL'], ENUM_RELOC_TYPE_ARM['R_ARM_THM_JUMP24']]:
            return thm_imm24_get(reloc_contents) + address + 4
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_THM_JUMP19']:
            return thm_imm19_get(reloc_contents) + address + 4
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_THM_JUMP11']:
            ins = self.builder.read_vaddr_typed(self.reloc.addr, '<H')
            if ins is None:
                return None
            return B_T1_get(ins) + address + 4
        if ty == ENUM_RELOC_TYPE_ARM['R_ARM_THM_JUMP8']:
            ins = self.builder.read_vaddr_typed(self.reloc.addr, '<H')
            if ins is None:
                return None
            return B_T2_get(ins) + address + 4
        if ty in [ENUM_RELOC_TYPE_ARM['R_ARM_MOVW_ABS_NC'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_THM_MOVW_ABS_NC'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_MOVT_ABS'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_THM_MOVT_ABS']]:
            return self.reloc.symbol + self.reloc.addend
        if ty in [ENUM_RELOC_TYPE_ARM['R_ARM_MOVW_PREL_NC'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_MOVT_PREL'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_THM_MOVW_PREL_NC'],
                  ENUM_RELOC_TYPE_ARM['R_ARM_THM_MOVT_PREL']]:
            return self.reloc.symbol

        return None


class RelocationAArch64(Relocation):
    def __init__(self, reloc, builder):
        super().__init__(reloc, builder)

    def get_target_ptr(self):
        return None


class RelocationX86(Relocation):
    def __init__(self, reloc, builder):
        super().__init__(reloc, builder)

    def get_target_ptr(self):
        ty = self.reloc.type

        if ty == ENUM_RELOC_TYPE_i386['R_386_32']:
            return self.builder.read_vaddr_typed(self.reloc.addr, '<I')

        elif (ty == ENUM_RELOC_TYPE_i386['R_386_GOT32'] or
              ty == 43): # R_386_GOT32X
            if self._is_patched_got32():
                return self.builder.read_vaddr_typed(self.reloc.addr, '<I')
            elif ty == 43 and self._is_patched_tls_get_addr_call():
                return None
            elif self.builder.got_address is None:
                return None
            else:
                delta = self.builder.read_vaddr_typed(self.reloc.addr, '<i')
                if delta is None:
                    return None
                return self.builder.got_address + delta

        elif ty == ENUM_RELOC_TYPE_i386['R_386_GOTOFF']:
            if self.builder.got_address is None:
                return None
            else:
                delta = self.builder.read_vaddr_typed(self.reloc.addr, '<i')
                if delta is None:
                    return None
                return self.builder.got_address + delta

        elif (ty == ENUM_RELOC_TYPE_i386['R_386_PC32'] or
              ty == ENUM_RELOC_TYPE_i386['R_386_PLT32'] or
              ty == ENUM_RELOC_TYPE_i386['R_386_GOTPC']):
            if self._is_patched_tls_get_addr_call():
                return None
            else:
                delta = self.builder.read_vaddr_typed(self.reloc.addr, '<i')
                if delta is None:
                    return None
                return self.reloc.addr - self.reloc.addend + delta

        else:
            return None

    # These are ported from the C++ code, see
    # src/RandoLib/posix/x86/Relocation.cpp for details
    def _is_patched_got32(self):
        b2 = self.builder.read_vaddr(self.reloc.addr - 2)
        return b2 in [0xc7, 0xf7, 0x81]

    def _is_patched_tls_get_addr_call(self):
        expected8_1 = b'\x65\xA1\x00\x00\x00\x00\x81\xE8'
        expected8_2 = b'\x65\xA1\x00\x00\x00\x00\x03\x83'
        expected8_3 = b'\x65\xA1\x00\x00\x00\x00\x8D\xB6'
        expected7_1 = b'\x65\xA1\x00\x00\x00\x00\x2D'
        expected7_2 = b'\x65\xA1\x00\x00\x00\x00\x90'
        return ((self.builder.read_vaddr(self.reloc.addr - 8, 8)
                 in [expected8_1, expected8_2, expected8_3]) or
                (self.builder.read_vaddr(self.reloc.addr - 7, 7)
                 in [expected7_1, expected7_2]))

class RelocationX64(Relocation):
    def __init__(self, reloc, builder):
        super().__init__(reloc, builder)

    def get_target_ptr(self):
        ty = self.reloc.type

        def pcrel_reloc():
            delta = self.builder.read_vaddr_typed(self.reloc.addr, '<l')
            if delta is None:
                return None
            return self.reloc.addr - self.reloc.addend + delta

        if ty == ENUM_RELOC_TYPE_x64['R_X86_64_32'] or ty == ENUM_RELOC_TYPE_x64['R_X86_64_32S']:
            return self.builder.read_vaddr_typed(self.reloc.addr, '<L')
        elif ty == ENUM_RELOC_TYPE_x64['R_X86_64_64']:
            return self.builder.read_vaddr_typed(self.reloc.addr, '<Q')
        elif (ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOT64'] or
              ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOTOFF64']):
            if self.builder.got_address is None:
                return None
            else:
                delta = self.builder.read_vaddr_typed(self.reloc.addr, '<q')
                if delta is None:
                    return None
                return self.builder.got_address + delta

        elif (ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOTPCREL'] or
              ty == 41 or  # R_X86_64_GOTPCRELX
              ty == 42):   # R_X86_64_REX_GOTPCRELX
            if self._is_patched_gotpcrel():
                return self.builder.read_vaddr_typed(self.reloc.addr, '<L')
            elif (self.builder.read_vaddr(self.reloc.addr - 1) == 0xe9 and
                  self.builder.read_vaddr(self.reloc.addr + 3) == 0x90):
                return ((self.reloc.addr - 1) - self.reloc.addend +
                        struct.unpack('<L', self.builder.read_vaddr(self.reloc.addr - 1, 4)))[0]
            else:
                return pcrel_reloc()
        elif (ty == ENUM_RELOC_TYPE_x64['R_X86_64_PC32'] or
              ty == ENUM_RELOC_TYPE_x64['R_X86_64_PLT32'] or
              ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOTPC32']):
            if self._is_patched_tls_get_addr_call():
                return None
            else:
                return pcrel_reloc()
        elif (ty == ENUM_RELOC_TYPE_x64['R_X86_64_PC64'] or
              ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOTPCREL64'] or
              ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOTPC64']):
            delta = self.builder.read_vaddr_typed(self.reloc.addr, '<q')
            if delta is None:
                return None
            return self.reloc.addr - self.reloc.addend + delta
        elif (ty == ENUM_RELOC_TYPE_x64['R_X86_64_TLSGD'] or
              ty == ENUM_RELOC_TYPE_x64['R_X86_64_TLSLD']):
            if self._is_pcrel_tlsxd():
                return pcrel_reloc()
        elif ty == ENUM_RELOC_TYPE_x64['R_X86_64_GOTTPOFF']:
            if self._is_pcrel_gottpoff():
                return pcrel_reloc()
        elif ty == ENUM_RELOC_TYPE_x64['R_X86_64_TLSDESC']:
            if self._is_pcrel_gotpc_tlsdesc():
                return pcrel_reloc()

        return None

    def _is_patched_gotpcrel(self):
        # BFD-specific hack: BFD sometimes replaces instructions like
        # OP %reg, foo@GOTPCREL(%rip) with immediate address versions:
        # OP %reg, $foo
        at_ptr = self.reloc.addr
        return (self.reloc.addend == -4 and
                (self.builder.read_vaddr(at_ptr - 1) >> 6 == 0x3 and
                 (self.builder.read_vaddr(at_ptr - 2) == 0xc7 or
                  self.builder.read_vaddr(at_ptr - 2) == 0xf7 or
                  self.builder.read_vaddr(at_ptr - 2) == 0x81)))

    def _is_patched_tls_get_addr_call(self):
        # TLS GD-IE or GD-LE transformation in gold:
        # replaces a call to __tls_get_addr with a
        # RAX-relative LEA instruction
        expected = b'\x64\x48\x8B\x04\x25\x00\x00\x00\x00\x48\x8D\x80'
        return self.builder.read_vaddr(self.reloc.addr - 12, 12) == expected

    def _is_pcrel_tlsxd(self):
        expected = b'\x48\x8d\x3d'
        return self.builder.read_vaddr(self.reloc.addr - 3, 3) == expected

    def _is_pcrel_gottpoff(self):
        at_ptr = self.reloc.addr
        return ((self.builder.read_vaddr(at_ptr - 2) == 0x8b or        # MOV
                 self.builder.read_vaddr(at_ptr - 2) == 0x03) and      # ADD
                (self.builder.read_vaddr(at_ptr - 1) & 0xc7) == 0x05)  # RIP-relative

    def _is_pcrel_gotpc_tlsdesc(self):
        at_ptr = self.reloc.addr
        return (self.builder.read_vaddr(at_ptr - 3) == 0x48 and
                self.builder.read_vaddr(at_ptr - 1) == 0x05 and
                (self.builder.read_vaddr(at_ptr - 2) == 0x8d or
                 self.builder.read_vaddr(at_ptr - 2) == 0x8b))


class TrapV2Builder():
    def __init__(self, filename):
        self.parser = ElfTrapParser(filename, trapv2_padding=True)
        self.binary = open(filename, 'r+b')
        self.records = []
        self.relocs = []
        self.records_built = False
        self.elf_file = ELFFile(self.binary)
        self.got_address = None
        # self.got_address = self.elf_file.get_section_by_name('.symtab').get_symbol_by_name('_GLOBAL_OFFSET_TABLE_')
        # assert(self.got_address is not None)

    def build_records(self):
        self.build_functions()
        self.sort_functions()
        self.compute_function_sizes()
        # self.trim_gaps()
        self.remove_empty_functions()
        self.place_relocs()
        self.assign_reloc_symbol_indices()
        self.records_built = True

    def write_v2(self, out=None):
        assert self.parser.header.version == 1, 'File does not contain TRaPv1 metadata'
        assert self.records_built, 'Builder not ready to write records'

        max_size = None
        if out is None:
            # write to the .txtrp section of self.binary
            txtrp_section = self.elf_file.get_section_by_name('.txtrp')
            self.binary.seek(txtrp_section.header.sh_offset)
            max_size = txtrp_section.header.sh_size
            out = self.binary

        writer = TrapWriter(out, self.parser.target_arch, max_size)

        trapv2_size = (TrapHeader.size_v2() +
                       4 + len(self.records) * TrapRecordV2.size(writer) +
                       4 + len(self.relocs) * TrapRelocV2.size(writer))
        if trapv2_size > max_size:
            print('ERROR: TRaPv2 does not fit in .txtrp section, need {} more bytes'.format(trapv2_size - max_size), file=sys.stderr)
            return False

        header = copy(self.parser.header)

        if header.have_data_refs:
            print('Cannot build TRaPv2 metadata: TRaP info has data references, which are not supported by TRaPv2')
            return False
        if header.have_nonexec_relo:
            print('Cannot build TRaPv2 metadata: Nonexec relocs are not yet supported by the builder')
            return False

        header.version = 2
        header.presorted_records = True
        header.have_pointer_size = True
        header.prepended_vector_size = True
        header.have_sym_size = True
        header.have_data_refs = False
        header.have_nonexec_relo = False
        header.pcrel_addr = False
        header.write_v2(writer)

        writer.write_uint32(len(self.records))
        for record in self.records:
            record.write_v2(writer)

        writer.write_uint32(len(self.relocs))
        for reloc in self.relocs:
            reloc.write_v2(writer)

        if max_size is not None:
            writer.zero_rest()

        self.binary.close()

        return True

    def is_one_byte_nop(self, addr):
        if (self.parser.target_arch == TargetArch.ARM or
            self.parser.target_arch == TargetArch.AARCH64):
            return False
        else:  # x86
            self.binary.seek(addr)
            byte = self.binary.read(1)
            return byte == b'\x90' or byte == b'\xcc'

    def build_functions(self):
        # ExecSectionProcessor::build_functions
        for record in self.parser.records():
            self.relocs.extend(record.relocs)

            for symbol in record.symbols:
                ty = TrapRecordV2.Type.FUNCTION if self._in_text(symbol.addr) else TrapRecordV2.Type.NONEXECREGION
                assert not self.parser.header.have_sym_size, "Sized symbols not yet handled by TRaPv2 builder"
                size = None
                self.records.append(TrapRecordV2(symbol.addr, size,
                                                 symbol.alignment, ty))

            if record.padding_addr != 0 and record.padding_size > 0:
                self.records.append(TrapRecordV2(record.padding_addr,
                                                 record.padding_size, 0,
                                                 TrapRecordV2.Type.PADDING))
                gap_addr = record.padding_addr + record.padding_size
                gap_ty = TrapRecordV2.Type.GAP if self._in_text(gap_addr) else TrapRecordV2.Type.NONEXECREGION
                self.records.append(TrapRecordV2(gap_addr, None, 0, gap_ty))

    def sort_functions(self):
        # ExecSectionProcessor::sort_functions
        def record_sort(record):
            # From RandLib Function::sort_rank()
            # If we have multiple functions at the same address, then:
            # 1) All the non-sized ones must come first, so we can set their sizes to 0
            # 2) Non-sized gaps should precede non-sized TRaP functions, so the latter have priority when computing sizes
            # 3) The sized function with non-zero size must come last (if it exists)
            if record.size is None:
                rank = 1 if record.type == TrapRecordV2.Type.GAP else 2
            elif record.size == 0:
                rank = 3
            else:
                rank = 4
            return (record.addr, rank)
        self.records.sort(key=record_sort)

    def compute_function_sizes(self):
        # ExecSectionProcessor::compute_function_sizes
        for (i, record) in enumerate(self.records):
            if record.size is not None:
                # size is already set
                continue

            if i + 1 < len(self.records):
                next_start = self.records[i + 1].addr
            else:
                next_start = self.parser.text_end

            # Don't run off the end of the text section
            if self._in_text(record.addr) and not self._in_text(next_start):
                next_start = self.parser.text_end
            record.size = next_start - record.addr

    def trim_gaps(self):
        # ExecSectionProcessor::trim_gaps
        for record in self.records:
            if record.type != TrapRecordV2.Type.GAP:
                continue

            while record.size > 0 and self.is_one_byte_nop(record.addr):
                record.addr += 1
                record.size -= 1

    def remove_empty_functions(self):
        # ExecSectionProcessor::remove_empty_functions
        self.records = list(filter(lambda r: r.size > 0, self.records))

    def place_relocs(self):
        """Attach relocs to their respective source records"""
        self.relocs.sort(key=lambda r: r.addr)
        record_idx = 0
        for (i, reloc) in enumerate(self.relocs):
            dest_record = None
            while dest_record is None:
                if record_idx >= len(self.records):
                    assert record_idx == len(self.records), "record_idx ran off end of records list"
                    gap_ty = TrapRecordV2.Type.GAP if self._in_text(reloc.addr) else TrapRecordV2.Type.NONEXECREGION
                    if self.records[-1].type == gap_ty:
                        # expand gap at end
                        self.records[-1].size = reloc.addr - self.records[-1].addr
                        record_idx -= 1
                    else:
                        self.records.append(TrapRecordV2(reloc.addr, self.parser.word_size, 0, gap_ty))
                    dest_record = self.records[-1]
                if reloc.addr < self.records[record_idx].addr:
                    if record_idx == 0:
                        self.records.insert(0, TrapRecordV2(reloc.addr, 0, 0, TrapRecordV2.Type.GAP))
                        record_idx = 1
                    dest_record = self.records[record_idx - 1]
                elif reloc.addr < self.records[record_idx].addr + self.records[record_idx].size:
                    dest_record = self.records[record_idx]
                else:
                    record_idx += 1

            if dest_record.reloc_count == 0:
                dest_record.reloc_idx = i
            dest_record.reloc_count += 1

    def assign_reloc_symbol_indices(self):
        """Update relocs to V2 with symbol index instead of address"""

        addrs = [record.addr for record in self.records]

        def update_reloc(reloc):
            idx = -1
            addend = 0 if reloc.addend is None else reloc.addend
            if reloc.symbol is not None:
                idx = bisect_left(addrs, reloc.symbol)
                if idx > 0 and (idx == len(addrs) or addrs[idx] != reloc.symbol):
                    idx -= 1
                addend = reloc.symbol - addrs[idx]
                if reloc.addend is not None:
                    addend += reloc.addend

                # TODO: test this on a binary with symbol references
                print(reloc, "{:#x}".format(addrs[idx]), addend)
            else:
                reloc_ptr = self._get_reloc(reloc).get_target_ptr()
                if reloc_ptr is not None:
                    reloc_target = bisect_left(addrs, reloc_ptr)
                    if (reloc_target < len(addrs) and
                        self.records[reloc_target].addr <= reloc_ptr and
                        reloc_ptr - self.records[reloc_target].addr < self.records[reloc_target].size):
                        idx = reloc_target
            return TrapRelocV2(reloc.addr, reloc.type, idx, addend)

        self.relocs = list(map(update_reloc, self.relocs))

    def _in_text(self, addr):
        return addr >= self.parser.text_begin and addr < self.parser.text_end

    def _get_reloc(self, reloc):
        if self.parser.target_arch == TargetArch.ARM:
            return RelocationArm(reloc, self)
        elif self.parser.target_arch == TargetArch.AARCH64:
            return RelocationAArch64(reloc, self)
        elif self.parser.target_arch == TargetArch.X86:
            return RelocationX86(reloc, self)
        elif self.parser.target_arch == TargetArch.X86_64:
            return RelocationX64(reloc, self)
        else:
            raise Exception("Unknown architecture")

    def read_vaddr(self, vaddr, length=1):
        try:
            file_offset = next(self.elf_file.address_offsets(vaddr))
        except StopIteration:
            # We might get here if the address falls outside
            # any known section (or between sections)
            return None

        self.binary.seek(file_offset)
        buf = self.binary.read(length)
        if length == 1:
            return buf[0]
        else:
            return buf

    def read_vaddr_typed(self, vaddr, type_str):
        sz = struct.calcsize(type_str)
        vaddr_bytes = self.read_vaddr(vaddr, sz)
        if vaddr_bytes is None:
            return None

        values = struct.unpack(type_str, vaddr_bytes)
        return values[0]

    # not sure if we should keep the gap records that don't contain relocations


def main():
    builder = TrapV2Builder(sys.argv[1])
    builder.build_records()

    # for record in builder.records:
    #     relocs = builder.relocs[record.reloc_idx:record.reloc_idx + record.reloc_count]
    #     print("{:#x}".format(record.addr), record.type, record.size, relocs)

    if not builder.write_v2():
        print('Error writing TRaPv2', file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()
