#!/usr/bin/env python2
#
# Copyright (c) 2014-2015, The Regents of the University of California
# Copyright (c) 2015-2019 RunSafe Security Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the University of California nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import enum
import os
import sys
import struct
import subprocess
import platform


class TargetArch(enum.Enum):
    ARM = 0
    AARCH64 = 1
    X86 = 2
    X86_64 = 3


class TrapWriter:
    def __init__(self, out, arch, max_size=None):
        self.out = out
        self.arch = arch
        self.pos = 0
        self.max_size = max_size
        if arch == TargetArch.ARM or arch == TargetArch.X86:
            self.uintptr_t = '<L'
            self.intptr_t = '<l'
            self.uint32_t = '<L'
            self.word_size = 32
        elif arch == TargetArch.AARCH64 or arch == TargetArch.X86_64:
            self.uintptr_t = '<Q'
            self.intptr_t = '<q'
            self.uint32_t = '<L'
            self.word_size = 64
        else:
            raise Exception('Unknown architecture')

    def _write(self, fmt, x):
        self.pos += struct.calcsize(fmt)
        assert self.pos <= self.max_size
        self.out.write(struct.pack(fmt, x))

    def write_uintptr(self, x):
        self._write(self.uintptr_t, x)

    def write_intptr(self, x):
        self._write(self.intptr_t, x)

    def write_uint32(self, x):
        self._write(self.uint32_t, x)

    def write_byte(self, x):
        self._write('=B', x)

    def write_reloc_type_symbol(self, ty, symbol):
        self.write_uintptr(((ty & 0xff) | (symbol << 8)) & (2**self.word_size - 1))

    def zero_rest(self):
        while self.pos < self.max_size:
            self.out.write(b'\0')
            self.pos += 1


class TrapParser(object):
    def __init__(self, filename, verbose=False, trapv2_padding=False):
        self.filename = filename
        self.verbose = verbose
        self.trapv2_padding = trapv2_padding

        # default to the host word size, overridden when dealing with ELF files
        self.word_size = struct.calcsize("P") * 8

        # set for ELF files
        self.target_arch = None

        self.trap_address = None
        self.trap_data = b''
        self.trap_data_len = 0
        self.trap_data_pos = 0

        self.got_plt_address = 0

        self.count_records = 0

        self.reloc_has_extra_symbol = lambda reloc_type: False
        self.reloc_has_extra_addend = lambda reloc_type: False

        self.header = None

    def records(self):
        self._parse_header()

        if self.verbose:
            if self.trap_address is not None:
                print(".txtrp@%08x" % self.trap_address)
            print('word size: ' + str(self.word_size))
            print("Header - Version: %02x Flags: %06x" % (self.header.version, self.header.flags))

        for record in self._parse_records():
            yield record

    def _parse_header(self):
        assert self.trap_data_pos == 0
        self.header = TrapHeader(struct.unpack('<I', self.trap_data[0:4])[0])
        self.trap_data_pos = 4

        # Non-exec relocation information follows the header
        if self.header.have_nonexec_relo:
            rel_addr = 0
            while True:
                delta = self.get_trap_uleb128()
                rel_type = self.get_trap_uleb128()
                if delta == 0 and rel_type == 0:
                    break

                rel_symbol, rel_addend = 0, 0
                if self.reloc_has_extra_symbol(rel_type):
                    rel_symbol = self.get_trap_value('@P')
                if self.reloc_has_extra_addend(rel_type):
                    rel_addend = self.get_trap_sleb128()

                rel_addr += delta
                if self.verbose:
                    print("Rel[%d]@%x=%x+%d" % (rel_type, rel_addr, rel_symbol, rel_addend))

        if self.header.have_pointer_size:
            self.header.pointer_size = self.get_trap_uleb128()
            self.word_size = self.header.pointer_size

    def _parse_records(self):
        while self.trap_data_pos < self.trap_data_len:
            record = TrapRecord()
            first_sym_addr = self.get_trap_value('@P')
            first_sym = True
            prev_sym_addr = 0
            end_pos = None

            if self.header.prepended_vector_size:
                length = self.get_trap_uleb128()
                end_pos = self.trap_data_pos + length

            # When not using prepended_vector_sizes this is an infinite loop
            # that we explicitly break out of when a terminating element is
            # found.
            while end_pos is None or self.trap_data_pos < end_pos:
                symbol = self._parse_symbol(prev_sym_addr)

                if not self.header.prepended_vector_size and symbol.addr == prev_sym_addr and symbol.size == 0 and symbol.alignment == 0 and not first_sym:
                    break

                if first_sym:
                    first_sym = False
                    record.addr = first_sym_addr - symbol.addr
                    symbol.addr = first_sym_addr
                    self.count_records = self.count_records + 1
                    if self.verbose:
                        print("Section Address: %08x" % (record.addr))

                prev_sym_addr = symbol.addr
                record.add_symbol(symbol)
                if self.verbose:
                    print("  " + repr(symbol))

            if self.header.record_relo:
                prev_rel_addr = record.addr
                end_pos = None
                if self.header.prepended_vector_size:
                    length = self.get_trap_uleb128()
                    end_pos = self.trap_data_pos + length

                while end_pos is None or self.trap_data_pos < end_pos:
                    reloc = self._parse_reloc(prev_rel_addr)
                    if not self.header.prepended_vector_size and reloc.addr == prev_rel_addr and reloc.type == 0:
                        break

                    prev_rel_addr = reloc.addr
                    record.add_reloc(reloc)
                    if self.verbose:
                        print("  " + repr(reloc))

            assert not self.header.have_data_refs, 'Parsing data references is not yet supported'

            if self.header.record_padding:
                padding_offset = self.get_trap_uleb128()
                padding_size = self.get_trap_uleb128()
                record.set_padding(padding_offset, padding_size)
                if self.verbose:
                    print("  Padding[%d]@%x/%x" % (padding_size, padding_offset, record.addr + padding_offset))

            if self.trapv2_padding:
                trapv2_padding_size = self.get_trap_value('<L')
                self.trap_data_pos += trapv2_padding_size
                if self.verbose:
                    print("  TRaPv2 padding: {}".format(trapv2_padding_size))

            yield record

    def _parse_symbol(self, prev_addr):
        delta = self.get_trap_uleb128()
        size = 0
        alignment = 0
        if self.header.have_sym_size:
            size = self.get_trap_uleb128()
        if self.header.have_sym_alignment:
            alignment = self.get_trap_uleb128()
        return TrapSymbol(prev_addr + delta, size, alignment)

    def _parse_reloc(self, prev_addr):
        delta = self.get_trap_uleb128()
        type = self.get_trap_uleb128()

        symbol, addend = None, None
        if self.reloc_has_extra_symbol(type):
            symbol = self.get_trap_value('@P')
        if self.reloc_has_extra_addend(type):
            addend = self.get_trap_sleb128()
        return TrapReloc(prev_addr + delta, type, symbol, addend)

    def get_trap_value(self, format):
        return_pcrel = False
        if format == '@P' and (self.header.pcrel_addr or self.header.baserel_addr):
            return_pcrel = True
            format = '=l' if self.word_size == 32 else '=q'
        elif format == '@P' and self.word_size == 32:
            format = '=L'
        elif format == '@P' and self.word_size == 64:
            format = '=Q'
        # We hijack '@S' here for signed ptrdiff_t
        elif format == '@S' and self.word_size == 32:
            format = '=l'
        elif format == '@S' and self.word_size == 64:
            format = '=q'

        size = struct.calcsize(format)
        value_raw = self.trap_data[self.trap_data_pos:self.trap_data_pos + size]
        value = struct.unpack(format, value_raw)[0]
        if return_pcrel:
            if self.header.baserel_addr:
                # We use GOTOFF relocs for most arches
                value += self.got_plt_address
            else:
                value += self.trap_address + self.trap_data_pos
        self.trap_data_pos += size
        return value

    def get_trap_uleb128(self):
        uleb_byte = self.get_trap_value('<B')
        uleb_shift = 0
        uleb = (uleb_byte & 0x7F)
        while (uleb_byte & 0x80) != 0:
            uleb_byte = self.get_trap_value('<B')
            uleb_shift += 7
            uleb |= ((uleb_byte & 0x7F) << uleb_shift)
        return uleb

    def get_trap_sleb128(self):
        sleb_byte = self.get_trap_value('<B')
        sleb_shift = 7
        sleb = (sleb_byte & 0x7F)
        while (sleb_byte & 0x80) != 0:
            sleb_byte = self.get_trap_value('<B')
            sleb |= ((sleb_byte & 0x7F) << sleb_shift)
            sleb_shift += 7

        if (sleb & (1 << (sleb_shift - 1))) != 0:
            sleb -= 1 << sleb_shift
        return sleb


class PeTrapParser(TrapParser):
    def __init__(self, filename, verbose=False, trapv2_padding=False):
        super(PeTrapParser, self).__init__(filename, verbose, trapv2_padding)
        import pefile
        infile = pefile.PE(self.filename)
        trap_list = [sec for sec in infile.sections if sec.Name[:6] == ".txtrp"]
        if len(trap_list) == 0:
            print("PE file does not contain a .txtrp section")
            sys.exit(1)

        trap_sec = trap_list[0]
        self.trap_data = trap_sec.get_data()
        self.trap_data_len = min(trap_sec.SizeOfRawData, trap_sec.Misc_VirtualSize)
        # print repr(trap_sec.get_data())

        if infile.FILE_HEADER.Machine == pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_AMD64']:
            self.word_size = 64


class ElfTrapParser(TrapParser):
    def __init__(self, filename, verbose=False, trapv2_padding=False):
        super(ElfTrapParser, self).__init__(filename, verbose, trapv2_padding)

        readelf_path = 'readelf'
        if 'SELFRANDO_BINUTILS_PREFIX' in os.environ:
            readelf_path = os.environ['SELFRANDO_BINUTILS_PREFIX'] + 'readelf'

        for line in subprocess.check_output([readelf_path, '-h', self.filename]).splitlines():
            if b'ELF32' in line:
                self.word_size = 32
            if b'ELF64' in line:
                self.word_size = 64

            if b'Machine:' in line:
                target_arch = line.split()[1]
                if target_arch == b'ARM':
                    self.target_arch = TargetArch.ARM
                elif target_arch == b'AArch64':
                    self.target_arch = TargetArch.AARCH64
                elif target_arch == b'Intel':  # Should actually be "Intel 80386"
                    self.target_arch = TargetArch.X86
                elif target_arch == b'Advanced':  # Should actually be "Advanced Micro Devices X86-64"
                    self.target_arch = TargetArch.X86_64
                else:
                    raise Exception("Unknown target architecture")

        for line in subprocess.check_output([readelf_path, '-x', '.txtrp', self.filename]).splitlines():
            if line.startswith(b'  0x'):
                self.trap_data += bytearray.fromhex(line[13:48].decode('utf-8'))
                if self.trap_address is None:
                    self.trap_address = int(line[2:12], 16)
        # We need the address of .got.plt
        for line in subprocess.check_output([readelf_path, '-d', self.filename]).splitlines():
            line_elems = line.split()
            if len(line_elems) >= 3 and line_elems[1] == "(PLTGOT)":
                self.got_plt_address = int(line_elems[2], 16)
                print(".got.plt@%08x" % self.got_plt_address)

        # get bounds of .text
        for line in subprocess.check_output([readelf_path, '-SW', self.filename]).splitlines():
            line_elems = line.split()
            if len(line_elems) < 6:
                continue

            if line_elems[1] == b".text":
                # Double-digit section index [NN]
                # takes only one spot in the elements
                text_size = int(line_elems[5], 16)
                self.text_begin = int(line_elems[3], 16)
                self.text_end = self.text_begin + text_size
            elif line_elems[2] == b".text":
                # Single-digit section index [ N] takes 2 spots
                text_size = int(line_elems[6], 16)
                self.text_begin = int(line_elems[4], 16)
                self.text_end = self.text_begin + text_size

        for line in subprocess.check_output([readelf_path, '-s', self.filename]).splitlines():
            if line.endswith(b'_TRaP_trap_begin'):
                line_elems = line.split()
                self.trap_begin = int(line_elems[1], 16)
            if line.endswith(b'_TRaP_trap_end'):
                line_elems = line.split()
                self.trap_end = int(line_elems[1], 16)

        self.trap_data_len = self.trap_end - self.trap_begin

        if self.target_arch == TargetArch.ARM:
            self.reloc_has_extra_symbol = lambda reloc_type: reloc_type in [
                43, 44, 45, 46, 47, 48, 49, 50]
            self.reloc_has_extra_addend = lambda reloc_type: reloc_type in [
                3, 24, 25, 26, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 96, 104, 105, 107]
        elif self.target_arch == TargetArch.AARCH64:
            self.reloc_has_extra_symbol = lambda reloc_type: reloc_type in [
                263, 264, 265, 266, 267, 268, 269, 275, 276, 277, 278, 284, 285, 286, 299]
            self.reloc_has_extra_addend = lambda reloc_type: reloc_type in [
                260, 261, 263, 264, 265, 266, 267, 268, 269, 275, 276, 277, 278, 284, 285,
                286, 299]
        elif self.target_arch == TargetArch.X86:
            self.reloc_has_extra_symbol = lambda reloc_type: False
            self.reloc_has_extra_addend = lambda reloc_type: reloc_type in [2, 4, 10]
        elif self.target_arch == TargetArch.X86_64:
            self.reloc_has_extra_symbol = lambda reloc_type: False
            self.reloc_has_extra_addend = lambda reloc_type: reloc_type in [
                2, 3, 4, 9, 19, 20, 22, 24, 26, 27, 28, 29, 30, 34, 41, 42]
        else:
            raise Exception("Unknown target architecture")

    def dump_to_file(self, filename):
        text_file = open(filename, "w")
        text_file.write(self.trap_data)
        text_file.close()


class TrapHeader(object):
    FLAG_BITS = {
        'function_beginnings_marked': 0,
        'presorted_records': 1,
        'have_sym_size': 2,
        'have_data_refs': 3,
        'record_relo': 4,
        'have_nonexec_relo': 5,
        'record_padding': 6,
        'pcrel_addr': 7,
        'have_sym_alignment': 8,
        'have_pointer_size': 9,
        'baserel_addr': 10,
        'prepended_vector_size': 11,
    }

    def __init__(self, value):
        self.version = value & 0xFF
        self.flags = value >> 8

    def __getattr__(self, name):
        if name in self.FLAG_BITS:
            return (self.flags & (1 << self.FLAG_BITS[name])) != 0
        else:
            raise(AttributeError('Could not find attribute: {}'.format(name)))

    def __setattr__(self, name, value):
        if name in self.FLAG_BITS:
            bit = 1 << self.FLAG_BITS[name]
            if value:
                self.flags |= bit
            else:
                self.flags &= ~bit
        else:
            super(TrapHeader, self).__setattr__(name, value)

    def write_v2(self, writer):
        value = self.version & 0xFF
        value |= self.flags << 8
        writer.write_uint32(value)
        writer.write_uint32(self.pointer_size)

    @staticmethod
    def size_v2():
        return 8


class TrapSymbol:
    def __init__(self, addr, size=0, alignment=0):
        self.addr = addr
        self.size = size
        self.alignment = alignment

    def __repr__(self):
        return "Sym@%x[%x] align:2^%d" % (self.addr, self.size, self.alignment)


class TrapReloc:
    def __init__(self, addr, type, symbol=None, addend=None):
        self.addr = addr
        self.type = type
        self.symbol = symbol
        self.addend = addend

    def __repr__(self):
        s = "Rel[{}]@{:#x}=".format(self.type, self.addr)
        if self.symbol is not None:
            s += "{:#x}".format(self.symbol)
        if self.addend is not None:
            s += '+{:#x}'.format(self.addend)
        return s


class TrapRecord:
    def __init__(self):
        self.addr = 0
        self.symbols = []
        self.relocs = []
        self.padding_addr = 0
        self.padding_size = 0

    def add_symbol(self, symbol):
        self.symbols.append(symbol)

    def add_reloc(self, reloc):
        self.relocs.append(reloc)

    def set_padding(self, offset, size):
        self.padding_addr = self.addr + offset
        self.padding_size = size


class TrapRecordV2:
    class Type(enum.IntEnum):
        FUNCTION = 0
        PADDING = 1
        GAP = 2
        NONEXECREGION = 3
        GOT = 4

    def __init__(self, addr, size, alignment, type):
        self.addr = addr
        self.size = size
        self.alignment = alignment
        self.type = type
        self.reloc_count = 0
        self.reloc_idx = 0

    def write_v2(self, writer):
        writer.write_uintptr(self.addr)
        writer.write_uintptr(self.size)
        writer.write_byte((self.alignment & 0x0f) | ((self.type & 0x0f) << 4))
        writer.write_uint32(self.reloc_count)
        writer.write_uint32(self.reloc_idx)

    @staticmethod
    def size(writer):
        return struct.calcsize(writer.uintptr_t) * 2 + 1 + 4 + 4


class TrapRelocV2:
    def __init__(self, addr, type, symbol_idx=None, addend=None):
        self.addr = addr
        self.type = type
        self.symbol_idx = symbol_idx
        self.addend = addend

    def __repr__(self):
        s = "Rel[{}]@{:#x}=".format(self.type, self.addr)
        if self.symbol_idx is not None:
            s += "{:#x}".format(self.symbol_idx)
        if self.addend is not None:
            s += '+{:#x}'.format(self.addend)
        return s

    def write_v2(self, writer):
        writer.write_uintptr(self.addr)
        writer.write_reloc_type_symbol(self.type, self.symbol_idx)
        writer.write_intptr(self.addend)

    @staticmethod
    def size(writer):
        return struct.calcsize(writer.uintptr_t) * 2 + struct.calcsize(writer.intptr_t)


def main():
    if platform.system() == 'Windows':
        parser = PeTrapParser(sys.argv[1], True)
    else:
        parser = ElfTrapParser(sys.argv[1], True)
    for record in parser.records():
        pass
    print("Finished parsing TRaP data")


if __name__ == "__main__":
    main()
